/*
  column: extracts selected columns from a NoSQL table.

  Copyright (c) 1998,1999,2000,2001,2002,2003 Carlo Strozzi

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  $Id: column.c,v 1.1 2003/04/19 20:35:35 carlo Exp $

*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/file.h>
#include <string.h>

#ifndef HELPFILE
#define HELPFILE "/usr/local/nosql/help/column.txt"
#endif

/* malloc() block size, in bytes. Do not set to less than sizeof(int) */
#define ALLOCSIZ 32

/* max. column name length is 31 (plus the trailing zero) */
#define MAXCOLNAME 31

typedef struct Header {
  char *names,			/* column names */
       *cmd,			/* buffer for optional back-end cmd */
       *args,			/* h.argv[] buffer */
       **argv;			/* h.argv[] for execvp(3) */
  int  top,			/* top used location in *names (0-n) */
       end,			/* last available position in of *names */
       *offset,			/* element offsets in *names */
       *saved,			/* saved *offset values */
       otop,			/* top used location in *offset */
       oend,			/* last available position in *offset */
       *seen,			/* *names elements selected for output */
       stop,			/* top used location in *seen */
       send,			/* last available position in *seen */
       ctop,			/* top used location in *cmd (0-n) */
       cend,			/* last available position in *cmd */
       atop,			/* top used location in *args (0-n) */
       aend,			/* last available position in *args */
       vtop,			/* top used location in h.argv[] */
       vend,			/* last available position in h.argv[] */
       argc;			/* argc for h.argv[] */
} Header;

struct {
  unsigned int header : 1;
  unsigned int debug : 1;
  unsigned int last : 1;
} flags;

/* global variables */
static char *progname;		/* global pointer to argv[0] */

/* Function declarations */

/* Unbuffered input routine */
static int getch(void) {
  char c;
  return (read(0, &c, 1) == 1) ? (unsigned char) c : EOF;
}

static int eusage(void) {
  fprintf(stderr, "Usage: %s [options] column-list'\n", progname);
  return 1;
}

static int etoolong(void) {
  fprintf(stderr, "%s: max. column name width (%d) exceeded\n",\
        progname,MAXCOLNAME-1);
  return 1;
}

static int addcmd(Header *h, const char *cmd, int len) {

  static int init=1;
  int l;

  /* init h->cmd if first time */

  if (init) {
     h->ctop = -1;
     h->cend = -1;

     if ((h->cmd = (char *)malloc(h->cend+=ALLOCSIZ)) == NULL) {
	perror(progname);
	exit(1);
     }
     h->cmd[0] = '\0';			/* init string (mandatory!) */
     init = 0;
  }

  l = strlen(cmd);

  if (len <= 0 || len > l) len = l;		/* set default */

  /* use '>=' in comparison, to account for string terminator */
  if ((h->ctop+len) >= h->cend) { /* realloc needed */
     if ((h->cmd = (char *)realloc(h->cmd,h->cend+=ALLOCSIZ)) == NULL) {
	perror(progname);
	exit(1);
     }
  }

  strncat(h->cmd, cmd, len);		/* append passed string to cmd */
  h->ctop += len;
  return len;
}

static int colpos(Header *h, char *name) {

  int i;

  if (flags.last) {
    for (i = h->otop; i >= 0; i--)
      if (!strcmp(name, h->names + h->offset[i]) && h->offset[i] >= 0)
	return i;
  }
  else {
    for (i = 0; i <= h->otop; i++)
      if (!strcmp(name, h->names + h->offset[i]) && h->offset[i] >= 0)
	return i;
  }

  return -1;
}

static int oldpos(Header *h, char *name) {

  int i;

  if (flags.last) {
    for (i = h->otop; i >= 0; i--)
      if (!strcmp(name, h->names + h->offset[i]) && h->offset[i] >= 0)
	return i;
  }
  else {
    for (i = 0; i <= h->otop; i++)
      if (!strcmp(name, h->names + h->saved[i]) && h->saved[i] >= 0)
	return i;
  }

  return -1;
}

static int seenpos(Header *h, char *name) {

  int i;

  if (flags.last) {
    for (i = h->stop; i >= 0; i--)
      if (!strcmp(name, h->names + h->seen[i]) && h->seen[i] >= 0)
	return i;
  }
  else {
    for (i = 0; i <= h->stop; i++)
      if (!strcmp(name, h->names + h->seen[i]) && h->seen[i] >= 0)
	return i;
  }

  return -1;
}

/* add column to output list */

static void addseen(Header *h, char *name) {

  static int init=1;

  /* init header if first time */

  if (init) {
     h->stop = -1;
     h->send = -1;

     if ((h->seen = (int *)malloc(ALLOCSIZ)) == NULL) {
	perror(progname);
	exit(1);
     }
     h->send += ALLOCSIZ/sizeof(int);
     init = 0;
  }

  /* use '>=' in comparison, to account for string terminator */
  if (h->stop >= h->send) {		/* realloc needed */
     if ((h->seen = (int *)realloc(h->seen,ALLOCSIZ)) == NULL) {
	perror(progname);
	exit(1);
     }
     h->send += ALLOCSIZ/sizeof(int);
  }

  /* add column offset for the new column, relative to h->names */
  *(h->seen + ++h->stop) = *(h->offset + colpos(h, name));
}

/* Add new column to the table header */

static void addcol(Header *h, char *name) {

  static int init=1;

  /* init header if first time */

  if (init) {
     h->top = -1;
     h->end = -1;
     h->otop = -1;
     h->oend = -1;

     if ((h->names = (char *)malloc(h->end+=ALLOCSIZ)) == NULL ||
	  (h->offset = (int *)malloc(ALLOCSIZ)) == NULL ||
	  (h->saved = (int *)malloc(ALLOCSIZ)) == NULL) {
	perror(progname);
	exit(1);
     }
     h->oend += ALLOCSIZ/sizeof(int);
     h->names[0] = '\0';			/* init string */
     init = 0;
  }

  /* ignore duplicated column names, if any. This is currently
     commented-out as duplicates are discarded anyway because of
     the overall logic of this program code */
  /* if (colpos(h, name) >= 0) return; */

  /* use '>=' in comparison, to account for string terminator */
  if ((h->top+strlen(name)) >= h->end) { /* realloc needed */
     if ((h->names = (char *)realloc(h->names,h->end+=ALLOCSIZ)) == NULL) {
	perror(progname);
	exit(1);
     }
  }

  /* add new column offset to the relevant array */
  if (h->otop >= h->oend) {		/* realloc needed */
     if ((h->offset = (int *)realloc(h->offset,ALLOCSIZ)) == NULL ||
	 (h->saved = (int *)realloc(h->saved,ALLOCSIZ)) == NULL) {
	perror(progname);
	exit(1);
     }
     h->oend += ALLOCSIZ/sizeof(int);
  }

  /* add offset for the new column, relative to h->names */
  *(h->offset + ++h->otop) = (h->top + 1);

  /* add new name to column names */
  h->top += (sprintf(h->names + h->top + 1, "%s", name) + 1);
}

/* insert new column to the left of a specified position */

static void inscol(Header *h, char *name, int pos) {

  int i, j;

  addcol(h, name);		/* append new column at the end */

  if (pos >= (h->otop - 1)) return;	/* done */

  j = *(h->offset + h->otop);		/* save rightmost value */

  /* right-shift offsets */

  for (i = h->otop; i > pos; i--)
	*(h->offset + i) = *(h->offset + i - 1);

  *(h->offset + i) = j;
}

/* add back-end cmd arguments to the execlp(3) argument array */

static void printhdr(Header *h) {

  int i, j;

  if (!flags.header) return;

   for (i=j=0; i <= h->stop; i++) {
      if (h->seen[i] >= 0) {

	 /* prepend TAB if not first column */
	 if (j++) printf("\t\001");
	 else printf("\001");

	 printf("%s", h->names + h->seen[i]);
      }
   }

   if (j) printf("\n");		/* add NL if at least one column */
}

int main(int argc, char *argv[]) {

  int i=0, j=0, k=0, x=0;
  char c;

  char tmpbuf[MAXCOLNAME] = "";	 	/* local work buffer */

  Header h;

  FILE *ifile = NULL, *ofile = NULL;

  const char *awk[] = {
      "BEGIN{FS=OFS=\"\\t\";}{print ",
      "}"
  };

  flags.header = 1; 		/* default it to print the header */
  flags.last = 0; 		/* default it to pick 1st occurrence */

  progname = argv[0];

  while (++i < argc && *argv[i] == '-') {

    if (!strcmp(argv[i], "-x") ||
  	     !strcmp(argv[i], "--debug")) flags.debug = 1;

    else if (!strcmp(argv[i], "-N") ||
  	     !strcmp(argv[i], "--no-header")) flags.header = 0;

    else if (!strcmp(argv[i], "-l") ||
  	     !strcmp(argv[i], "--last")) flags.last = 1;

    else if (!strcmp(argv[i], "-i") ||
    	     !strcmp(argv[i], "--input")) {

      if (++i >= argc || *argv[i] == '-') exit(eusage());

      if ((ifile = freopen(argv[i], "r", stdin)) < 0) {
         perror(argv[i]);
         exit(1);
      }
    }

    else if (!strcmp(argv[i], "-o") ||
    	     !strcmp(argv[i], "--output")) {

      if (++i >= argc || *argv[i] == '-') exit(eusage());

      if ((ofile = freopen(argv[i], "w", stdout)) < 0) {
    	 perror(argv[i]);
    	 exit(1);
      }
    }

    else if (!strcmp(argv[i], "-h") ||
    	     !strcmp(argv[i], "--help")) {

      execlp("grep","grep","-v","^#",HELPFILE,(char *) 0);
      perror("grep");
      exit(1);
    }
  }

  while ((c = getch()) != EOF) {

     if (c == '\001') continue;			/* ignore SOH chars */

     if (j >= MAXCOLNAME) exit(etoolong());

     if (c != '\t' && c != '\n') {
        tmpbuf[j++] = c;
	continue;
     }

     tmpbuf[j++] = '\0';			/* set terminator */
     addcol(&h, tmpbuf);			/* append to header */
     j = 0;

     if (c == '\n') {

	/* save input column positions before moving them around */
	for (k=0; k<= h.otop; k++)
		*(h.saved + k) = *(h.offset + k);

	addcmd(&h, awk[0], 0);			/* init AWK program */

	h.stop = h.send = -1;			/* mandatory, or segfault */

	while (i < argc) {
	  if ((k=oldpos(&h, argv[i])) >= 0)
	     x = k+1;				/* save insertion point */
	  else {				/* insert column if new */
	    inscol(&h, argv[i], x++);
	    k = h.otop;
	  }

	  /* make sure h.stop and h.send are initialized before calling
	     seenpos(), or the function may segfault. See above. */

	  if (seenpos(&h, argv[i]) < 0) {

	     /* prepend comma if not first column */
	     if (j++) sprintf(tmpbuf, ",$(%d)", k+1);
	     else sprintf(tmpbuf, "$(%d)", k+1);

	     addcmd(&h, tmpbuf, 0);
	     addseen(&h, argv[i]);		/* add to the seen list */
	  }
	  i++;
	}

	addcmd(&h, awk[1], 0);

	if (flags.debug) fprintf(stderr, "mawk %s\n", h.cmd);

	if (!j) exit(0);	/* exit if no columns were selected */

	printhdr(&h);			/* print header to stdout */

        fflush(NULL);	/* Make sure the header is output */
        execlp("mawk", "mawk", h.cmd, (char *) 0);
        perror("mawk");
        exit(1);
     }
  }

  exit(0);			/* Not reached */
}

/* EOF */
